#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

MOUSE_PROTOCOL_NONE         = 0
MOUSE_PROTOCOL_X10          = 9
MOUSE_PROTOCOL_NORMAL       = 1000
MOUSE_PROTOCOL_HIGHLIGHT    = 1001
MOUSE_PROTOCOL_BUTTON_EVENT = 1002
MOUSE_PROTOCOL_ANY_EVENT    = 1003

MOUSE_ENCODING_NORMAL       = 1000
MOUSE_ENCODING_UTF8         = 1005
MOUSE_ENCODING_SGR          = 1006
MOUSE_ENCODING_URXVT        = 1015

FOCUS_EVENT_TRACKING        = 1004

